#include "srm_client.h"

#include "../../misc/log_time.h"
#include "../../misc/url_options.h"

#include <string>
#include <list>

class SRM22ClientTest {
  
private:
  SRMClientRequest* req;
  SRMClient* client;
public:
  SRM22ClientTest(std::string);
  SRM22ClientTest(std::list<std::string>);
  ~SRM22ClientTest();
  int test_ping();
  int test_getTURLs();
  int test_putTURLs(std::string space_token="");
  int test_info();
  int test_list_dir();
  std::list<std::string> test_list_tokens(std::string desc="");
  int test_list_request_tokens();
  int test_stage();
  int release_get_files();
  int release_put_files();
  int abort_request();

};

SRM22ClientTest::SRM22ClientTest(std::list<std::string> urls) {
  
  std::list<std::string> empty_urls;

  try {
    odlog(DEBUG)<<"Create request with empty list"<<std::endl;
    req = new SRMClientRequest(empty_urls);
  }
  catch (SRMInvalidRequestException& e) {
    odlog(DEBUG)<<"cannot pass empty list"<<std::endl;
  };

  try {
    odlog(DEBUG)<<"Create request correctly"<<std::endl;
    req = new SRMClientRequest(urls);
  }
  catch (SRMInvalidRequestException& e) {
    odlog(ERROR)<<"ERROR: This should not happen!"<<std::endl;
    throw e;
  };

  bool timeout = false;
  client = SRMClient::getInstance(urls.front(), timeout, "", 30);

};

SRM22ClientTest::SRM22ClientTest(std::string url) {

  req = new SRMClientRequest(url);
  bool timeout = false;
  client = SRMClient::getInstance(url, timeout, "", 30);
};

SRM22ClientTest::~SRM22ClientTest() {
  delete req;
  delete client;
};

int SRM22ClientTest::test_ping() {

  std::string version;

  if(client->ping(version) == SRM_OK) {
    odlog(DEBUG) << "Version is " << version << std::endl;
    return 0;
  }
  odlog(DEBUG) << "Error retrieving version with srmPing" << std::endl;
  return 1;
  
};


int SRM22ClientTest::test_getTURLs() {

  std::list<std::string> turls;

  client->getTURLs(*req, turls);

  for (std::list<std::string>::iterator i = turls.begin();
       i != turls.end();
       ++i) {
    odlog(DEBUG) << "TURL: " << *i << std::endl;
  };

  return 0;
};


int SRM22ClientTest::test_putTURLs(std::string space_token) {

  std::list<std::string> turls;

  // set space token if given in URL and not in parameters
  if (space_token.compare("") == 0) {
    if (get_url_option(req->surls().front(), "spacetoken", space_token)) {
      if(client->getVersion().compare("v2.2") == 0) {
        // only print message if using v2.2
        odlog(DEBUG)<<"No space token specified"<<std::endl;
      };
    }
    else {
      if(client->getVersion().compare("v2.2") != 0) {
        // print warning if not using srm2.2
        odlog(WARNING)<<"Warning: Using SRM protocol v1 which does not support space tokens"<<std::endl;
    }
      else {
        odlog(DEBUG)<<"Using space token "<<space_token<<std::endl;
      };
    };
  };

  // take out options in srm url
  std::string can_url(req->surls().front());
  if (canonic_url(can_url)) {
    odlog(ERROR) << "Error getting canonic url" << std::endl;
    return -1;
  };
      
  req = new SRMClientRequest(can_url);
  if(space_token.compare("") != 0) req->space_token(space_token);

  odlog(DEBUG) << "URL is now " << req->surls().front() << std::endl;

  // put 1MB file
  client->putTURLs(*req, turls, 1000000);

  for (std::list<std::string>::iterator i = turls.begin();
       i != turls.end();
       ++i) {
    odlog(DEBUG) << "TURL: " << *i << std::endl;
  };

  return 0;
};

int SRM22ClientTest::test_info() {

  odlog(DEBUG) << "SRM client version " << client->getVersion() << std::endl;

  std::list<struct SRMFileMetaData> metadata;

  client->info(*req, metadata);

  if(!metadata.empty()) {
    // size is a random int if not defined, so check path is defined
    if(metadata.front().path != "") {
      odlog(DEBUG) << "size: " << metadata.front().size << std::endl;
    };
    if(metadata.front().checkSumType != "" &&
       metadata.front().checkSumValue != "") {
      odlog(DEBUG) << "checksum: " << metadata.front().checkSumType << ":" << metadata.front().checkSumValue << std::endl;
    };
    std::string locality = "";
    if (metadata.front().fileLocality == SRM_ONLINE) {locality = "online";}
    if (metadata.front().fileLocality == SRM_NEARLINE) {locality = "nearline";}
    odlog(DEBUG) << "locality: " << locality << std::endl;
  };

  return 0;
};

int SRM22ClientTest::test_list_dir() {

  std::list<struct SRMFileMetaData> metadata;

  client->info(*req, metadata);

  for (std::list<struct SRMFileMetaData>::iterator i = metadata.begin();
       i != metadata.end();
       ++i) {

    // size is a random int if not defined, so check path is defined
    if((*i).path != "") {
      odlog(DEBUG) << "size: " << (*i).size << std::endl;
    };
    if((*i).checkSumType != "" &&
       (*i).checkSumValue != "") {
      odlog(DEBUG) << "checksum: " << (*i).checkSumType << ":" << (*i).checkSumValue << std::endl;
    };
    std::string locality = "";
    if ((*i).fileLocality == SRM_ONLINE) {locality = "online";}
    if ((*i).fileLocality == SRM_NEARLINE) {locality = "nearline";}
    odlog(DEBUG) << "locality: " << locality << std::endl;
  };

  return 0;
};

std::list<std::string> SRM22ClientTest::test_list_tokens(std::string desc) {

  std::list<std::string> tokens;

  if(client->getSpaceTokens(tokens, desc) != SRM_OK)
    odlog(DEBUG) << "Error listing space tokens" << std::endl;

  for (std::list<std::string>::iterator i = tokens.begin();
       i != tokens.end();
       ++i) {
    odlog(DEBUG) << "Token: " << *i << std::endl;
  };  

  return tokens;
};

int SRM22ClientTest::test_list_request_tokens() {

  std::list<std::string> tokens;
  std::string desc = "";

  char * user = getlogin();
  if (user != 0) {
    desc = "stage_" + std::string(user);
    odlog(DEBUG)<<"userRequestDescription is "<<desc<<std::endl;
  };

  if(client->getRequestTokens(tokens, desc) != SRM_OK) {
    odlog(DEBUG) << "Error listing request tokens" << std::endl;
    return -1;
  };

  for (std::list<std::string>::iterator i = tokens.begin();
       i != tokens.end();
       ++i) {
    odlog(DEBUG) << "Token: " << *i << std::endl;
  };  

  return 0;
};

int SRM22ClientTest::test_stage() {

  odlog(DEBUG)<<"Testing stage"<<std::endl;

  if (client->requestBringOnline(*req) != SRM_OK) {
    odlog(ERROR)<<"Error in submitting bring online request"<<std::endl;
    return 1;
  };

  // check if all files are online
  if (req->status() == SRM_REQUEST_FINISHED_SUCCESS) {
    odlog(DEBUG)<<"All files are already online"<<std::endl;
    return 0;
  };

  // query until all files are online
  do {
    int sleeptime = req->waiting_time();
    odlog(DEBUG)<<"sleeping "<<sleeptime<<" seconds"<<std::endl;
    sleep(sleeptime);

    if (client->requestBringOnlineStatus(*req) != SRM_OK) {
      if(req->status() == SRM_REQUEST_ONGOING) {
        odlog(ERROR)<<"Failed to query status of bring online, will retry"<<std::endl;
      } else {
        odlog(ERROR)<<"Failed to query status of bring online or all files failed"<<std::endl;
      }
    };
    std::map<std::string, SRMFileLocality> statuses = req->surl_statuses();
    for(std::map<std::string, SRMFileLocality>::iterator i = statuses.begin();
       i != statuses.end();
       ++i) {
      odlog(DEBUG)<<"File "<<(*i).first<<std::endl;
      SRMFileLocality file_locality = (*i).second;
      if (file_locality == SRM_ONLINE) {odlog(DEBUG)<<"Status ONLINE"<<std::endl;}
      else if (file_locality == SRM_NEARLINE) {odlog(DEBUG)<<"Status STAGING"<<std::endl;}
      else if (file_locality == SRM_STAGE_ERROR) {odlog(DEBUG)<<"Status ERROR"<<std::endl;}
      else {odlog(DEBUG)<<"Status UNKNOWN"<<std::endl;};
    };
  } while(req->status() == SRM_REQUEST_ONGOING);
  return 0;
};

int SRM22ClientTest::release_get_files() {

  if (client->releaseGet(*req)){
    odlog(DEBUG) << "Files released ok" << std::endl;
  }
  return 0;
};

int SRM22ClientTest::release_put_files() {

  if (client->releasePut(*req)){
    odlog(DEBUG) << "Files released ok" << std::endl;
  }
  return 0;
};

int SRM22ClientTest::abort_request() {

  if (client->abort(*req)){
    odlog(DEBUG) << "Request aborted ok" << std::endl;
  }
  return 0;
};


int main() {

  // set logging
  LogTime::Active(true);
  LogTime::Level(DEBUG);

  // test staging, with files already online
  std::list<std::string> surls;
  surls.clear();
  surls.push_back("srm://srm.ndgf.org/srm/managerv2/SFN=/pnfs/ndgf.org/data/atlas/tape/calib0/RDO/calib0.005011.J2_pythia_jetjet.digit.RDO.v12003101_tid003438/calib0.005011.J2_pythia_jetjet.digit.RDO.v12003101_tid003438._00162.pool.root.7__DQ2-1200659973");
  surls.push_back("srm://srm.ndgf.org/srm/managerv2/SFN=/pnfs/ndgf.org/data/atlas/tape/calib0/RDO/calib0.005011.J2_pythia_jetjet.digit.RDO.v12003101_tid003438/calib0.005011.J2_pythia_jetjet.digit.RDO.v12003101_tid003438._00152.pool.root.2");

  //surls.push_back("srm://srm-atlas.gridpp.rl.ac.uk/castor/ads.rl.ac.uk/prod/atlas/stripInput/atlasdatadisk/test/00/test.040208.CERN.DISK.00/040208-1730-CERN-DISK-TEST.1__DQ2-1202220904");

  //surls.push_back("srm://srm-atlas.cern.ch/castor/cern.ch/grid/atlas/dq2/misal1_mc12/RDO/misal1_mc12.005812.2loosefiltered_minbias7.digit.RDO.v12000605_tid018312/RDO.018312._69124.pool.root.2");

  //srm://srm-atlas.cern.ch/castor/cern.ch/grid/atlas/dcameron/test.210108.01");
  SRM22ClientTest * test = new SRM22ClientTest(surls);
  test->test_stage();

  test = new SRM22ClientTest(std::string("srm://srm.ndgf.org/srm/managerv2/SFN=/pnfs/ndgf.org/data/atlas/disk/user/user.mlassnig.dataset.1/dummyfile3"));

  //SRM22ClientTest* test = new SRM22ClientTest(std::string("srm://srm-atlas.cern.ch/castor/cern.ch/grid/atlas/dq2/misal1_mc12/RDO/misal1_mc12.005812.2loosefiltered_minbias7.digit.RDO.v12000605_tid018312/RDO.018312._69124.pool.root.2"));

  // Endpoint is https for storm so this doesn't work
  //SRM22ClientTest* test = new SRM22ClientTest(std::string("srm://storm-fe.cr.cnaf.infn.it:8444/atlas/datafiles/user/user.mlassnig.dataset.1/dummyfile3"));

  // test get and info
  test->test_ping();
  test->test_getTURLs();
  test->test_info();
  test->release_get_files();

  // test listing directory. use list of surls
  //std::list<std::string> surls;
  surls.push_back("srm://srm.ndgf.org/srm/managerv2/SFN=/pnfs/ndgf.org/data/atlas/disk/user/user.mlassnig.dataset.1");
  test = new SRM22ClientTest(surls);
  test->test_list_dir();

  // test put
  test = new SRM22ClientTest(std::string("srm://srm.ndgf.org/srm/managerv2/SFN=/pnfs/ndgf.org/data/atlas/disk/user/user.mlassnig.dataset.1/dummyfile3.test"));
  test->test_putTURLs();

  // list supported space tokens
  std::list<std::string> tokens = test->test_list_tokens("ATLASDATADISK");

  if (!tokens.empty()) {
    // test put with space tokens
    test = new SRM22ClientTest(std::string("srm://srm.ndgf.org/srm/managerv2/SFN=/pnfs/ndgf.org/data/atlas/disk/user/user.mlassnig.dataset.1/dummyfile3.test"));
    test->test_putTURLs(tokens.front());
  };

  // test with bad token - should give error
  test = new SRM22ClientTest(std::string("srm://srm.ndgf.org;spacetoken=BADTOKEN/srm/managerv2/SFN=/pnfs/ndgf.org/data/atlas/disk/user/user.mlassnig.dataset.1/dummyfile3.test"));
  test->test_putTURLs();


  // wait and then abort the request
  sleep(1);
  //test->release_put_files();
  test->abort_request();

  test->test_list_request_tokens();


  //test = new SRM22ClientTest(std::string("srm://gridka-dcache.fzk.de/pnfs/gridka.de/atlas/disk-only/user/user.mlassnig.dataset.1/dummyfile3"));

  delete test;
  return 0;
};
